/*
 * @file purple-fifo.c  A FIFO byte queue.
 * @ingroup core
 *
 * Pidgin is the legal property of its developers, whose names are too numerous
 * to list here.  Please refer to the COPYRIGHT file distributed with this
 * source distribution.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "internal.h"

#define BUFFER_SIZE 256

typedef struct
{
	GList *buffer_list;
	GList *enqueue_buffer;
	gint   enqueue_pos;
	GList *dequeue_buffer;
	gint   dequeue_pos;

	gint   total_used;
}
PurpleFifo;

PurpleFifo *
purple_fifo_new (void)
{
	return g_new0 (PurpleFifo, 1);
}

void
purple_fifo_destroy (PurpleFifo *fifo)
{
	GList *l;

	for (l = fifo->buffer_list; l; l = g_list_next (l))
		g_free (l->data);

	g_list_free (fifo->buffer_list);
}

gint
purple_fifo_get_used (PurpleFifo *fifo)
{
	return fifo->total_used;
}

static void
ensure_buffers (PurpleFifo *fifo)
{
	if (!fifo->buffer_list)
	{
		fifo->buffer_list = g_list_append (NULL, g_malloc (BUFFER_SIZE));
		fifo->enqueue_buffer = fifo->buffer_list;
		fifo->dequeue_buffer = fifo->buffer_list;

		fifo->buffer_list->data = g_malloc (BUFFER_SIZE);
	}
}

static void
enqueue_next_buffer (PurpleFifo *fifo)
{
	g_list_append (fifo->enqueue_buffer, g_malloc (BUFFER_SIZE));
	fifo->enqueue_buffer = g_list_next (fifo->enqueue_buffer);
	fifo->enqueue_pos = 0;
}

static void
dequeue_next_buffer (PurpleFifo *fifo)
{
	g_free (fifo->dequeue_buffer->data);
	fifo->dequeue_buffer = g_list_delete_link (fifo->dequeue_buffer, fifo->dequeue_buffer);

	fifo->buffer_list = fifo->dequeue_buffer;
	if (!fifo->buffer_list)
		fifo->enqueue_buffer = NULL;

	fifo->dequeue_pos = 0;
}

void
purple_fifo_enqueue (PurpleFifo *fifo, gconstpointer libpurple, gint n)
{
	gint in_pos;

	ensure_buffers (fifo);

	for (in_pos = 0; in_pos < n; )
	{
		gint n_copied = MIN (BUFFER_SIZE - fifo->enqueue_pos, n - in_pos);

		memcpy (((guint8 *) fifo->enqueue_buffer->data) + fifo->enqueue_pos, ((const guint8 *) libpurple) + in_pos, n_copied);

		fifo->enqueue_pos += n_copied;
		in_pos            += n_copied;

		if (fifo->enqueue_pos == BUFFER_SIZE)
			enqueue_next_buffer (fifo);
	}

	fifo->total_used += n;
}

gboolean
purple_fifo_dequeue (PurpleFifo *fifo, gpointer dest, gint n)
{
	gint out_pos;

	if (n > fifo->total_used)
		return FALSE;

	for (out_pos = 0; out_pos < n; )
	{
		gint n_copied = MIN (BUFFER_SIZE - fifo->dequeue_pos, n - out_pos);

		if (dest)
			memcpy (((guint8 *) dest) + out_pos,
				((guint8 *) fifo->dequeue_buffer->data) + fifo->dequeue_pos,
				n_copied);

		fifo->dequeue_pos += n_copied;
		out_pos           += n_copied;

		if (fifo->dequeue_pos == BUFFER_SIZE)
			dequeue_next_buffer (fifo);
	}

	fifo->total_used -= n;
	return TRUE;
}

gboolean
purple_fifo_dequeue_line (PurpleFifo *fifo, gchar **dest)
{
	gchar    *line;
	gboolean  have_cr   = FALSE;
	gboolean  have_crlf = FALSE;
	GList    *buffer_l;
	gint      pos;
	gint      n;

	buffer_l = fifo->dequeue_buffer;
	n = 0;

	for (pos = fifo->dequeue_pos; n < fifo->total_used && !have_crlf; pos++, n++)
	{
		guint8 *p = ((guint8 *) buffer_l->data) + fifo->dequeue_pos;
		guint8  c = *p;

		if (c == '\r')
		{
			have_cr   = TRUE;
			have_crlf = FALSE;
		}
		else if (c == '\n' && have_cr)
		{
			have_crlf = TRUE;
		}

		if (pos == BUFFER_SIZE)
		{
			buffer_l = g_list_next (buffer_l);
			pos = 0;
		}
	}

	if (!have_crlf)
		return FALSE;

	line = g_malloc (n + 1);
	if (!purple_fifo_dequeue (fifo, line, n))
	{
		/* Paranoid check; can't happen in practice */
		g_free (line);
		return FALSE;
	}

	*(line + n) = '\0';
	*dest = line;

	return TRUE;
}

gboolean
purple_fifo_peek (PurpleFifo *fifo, gpointer dest, gint n)
{
	GList *dequeue_buffer;
	gint   dequeue_pos;
	gint   out_pos;

	if (n > fifo->total_used)
		return FALSE;

	dequeue_buffer = fifo->dequeue_buffer;
	dequeue_pos    = fifo->dequeue_pos;

	for (out_pos = 0; out_pos < n; )
	{
		gint n_copied = MIN (BUFFER_SIZE - dequeue_pos, n - out_pos);

		if (dest)
			memcpy (((guint8 *) dest) + out_pos,
				((guint8 *) dequeue_buffer->data) + dequeue_pos,
				n_copied);

		dequeue_pos += n_copied;
		out_pos     += n_copied;

		if (dequeue_pos == BUFFER_SIZE)
		{
			dequeue_buffer = g_list_next (dequeue_buffer);
			dequeue_pos    = 0;
		}
	}

	return TRUE;
}
